/* $Id: e2_fs_walk.c 379 2007-05-05 07:51:22Z tpgww $
 *
 * Copyright (C) 2005-2007 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/filesystem/e2_fs_walk.c
 * @brief tree-walk functionality, like nftw() but with user-data and additional flags
 **/

#include "e2_fs.h"
#include <string.h>
#include <dirent.h>

/* twstat pointer to function to be used for stat oprations
 * (e2_fs_lstat() for E2TW_PHYS walks, e2_fs_stat() for others)
 * cbfunc pointer to the function to be called for each located item
 * user_data void pointer to user-specified data to be supplied to cbunc each
 *    time it's called (with GError** as its first member if E2TW_XERR is set)
 * max_depth the largest number of tree levels to descend, or -1 for no limit
 * exec_flags flags dictating the way the walk is to be performed
 * device the system on which the starting item resides
 * original_path newly-allocated localised path string, = cwd when the walk is started
 */
typedef struct _E2_TwData
{
	gint (*twstat) (const gchar *, struct stat *
#ifdef E2_VFS
		, GError **
#endif
	);
	E2_TwResult (*cbfunc) (const gchar *, const struct stat *, E2_TwStatus, void *);
	void *user_data;
	gint max_depth;
	E2_TwFlags exec_flags;
	dev_t device;
//	gchar *original_path;
} E2_TwData;

/**
 * @brief determine the appropriate status report code for link @a local
 *
 * With relative symlinks, stat() generally fails, and anyway, can't
 * distinguish valid links from hanging ones. Nor can lstat().
 * So this uses a kludgy but robust alternative approach
 *
 * @param local absolute path of link-item to process, localised string
 *
 * @return callbck status code in accordance with whether the link is valid
 **/
static E2_TwResult _e2_fs_tw_link (gchar *local)
{
	E2_TwStatus code;
	gchar buf[PATH_MAX];
	gint read = e2_fs_readlink (local, buf, sizeof (buf) E2_ERR_NONE());
	if (read == -1)
	{
		code = E2TW_SLN;
	}
	else
	{
		buf[read] = '\0';
		gchar *converted1 = F_FILENAME_FROM_LOCALE (local);
		gchar *converted2 = F_FILENAME_FROM_LOCALE (buf);
		gchar *targetpath = e2_utils_translate_relative_path (converted1, converted2);
		F_FREE (converted1);
		F_FREE (converted2);
		//targetpath has a trailing /, with which access() can't cope
		gint len = strlen (targetpath) - 1;
		*(targetpath+len) = '\0';

		converted1 = F_FILENAME_TO_LOCALE (targetpath);
		//access() needs to look through links
		code = (e2_fs_access (converted1, F_OK E2_ERR_NONE())) ? E2TW_SLN : E2TW_SL;
		g_free (targetpath);
		F_FREE (converted1);
	}
	return code;
}
/**
 * @brief process 'non-directory' item @a local, as part of a treewalk
 *
 * This determines the appropriate status report code for the callback,
 * then calls it
 * @a local may be the path of a link, regular file, etc
 * @a local will have been stat'd or lstat'd (depending on walk-flag
 * E2TW_PHYS) into @a statbuf, before this function is called
 *
 * @param local absolute path of item to process, localised string
 * @param statbuf pointer to struct stat used for stating @a local
 * @param twdata pointer to struct with 'constant' data for each call here
 *
 * @return result code in accordance with callback-function response
 **/
static E2_TwResult _e2_fs_tw_nondir (gchar *local, struct stat *statbuf,
	E2_TwData *twdata)
{
	E2_TwStatus code;
	struct stat statbuf2;	//don't disturb statbuf
	if (!(twdata->exec_flags & E2TW_PHYS)
	//the prior stat() processed a non-link, or might, or might not,
	//have successfully processed a link
	//check if it was a link
		&& !e2_fs_lstat (local, &statbuf2 E2_ERR_NONE()) 	//stat link succeeded
		&& S_ISLNK (statbuf2.st_mode))
			code = _e2_fs_tw_link (local);

	else	//the prior lstat did not look through a link
			//or the local lstat failed
			//in general, relative links (valid or not) will end up here
		if (S_ISLNK (statbuf->st_mode))
			code = _e2_fs_tw_link (local);
	else
		code = E2TW_F;

	E2_TwResult result = (*twdata->cbfunc) (local, statbuf, code, twdata->user_data);
	return result;
}
/**
 * @brief process directory @a local, as part of a treewalk
 *
 * if possible, and in accord with the walk-mode determined by flags in @a twdata,
 * @a local is opened and its contents are each 'reported' to the callback
 * function recorded in @a twdata.
 * Any subdir found among the contents will recurse to this
 * function.
 * Non-dirs are handled by another function, before coming
 * here or when found.
 * As appropriate, status report code(s) provided to the specified
 * callback function will be:
 * E2TW_F item is not a directory or link
 * E2TW_SL symbolic link to a file other than a directory
 * E2TW_SLN symbolic link whose target file is non-existent
 * E2TW_D	directory
 * E2TW_DL directory, not opened due to tree-depth limit
 * E2TW_DM directory, not opened due to different file system
 * E2TW_DNR unreadable directory
 * E2TW_DP directory processed (all its contents have been reported)
 * E2TW_NS un-statable item
 *
 * E2TW_D and E2TW_DP will both be reported for any opened
 * directory (unless, of course, processing is aborted by a message
 * from the callback, between the two reports)
 *
 * @param local absolute path of _directory_ to process, localised string, with or without trailing "/"
 * @param depth enumator of current parent's level in the tree (relative
 *      to the original parent), 0 to max_depth (if the latter is not -1)
 * @param twdata pointer to struct with 'constant' data for each call here
 *
 * @return result code in accordance with callback-function responses, E2TW_STOP or E2TW_CONTINUE if all was well
 **/
 static E2_TwResult _e2_fs_tw (const gchar *local, gint depth, E2_TwData *twdata)
{
	struct stat statbuf;
	E2_TwResult result;
	GList *entries, *parked_dirs, *tmp;
	gint newdepth;
#ifdef E2_VFS
	GError **callerr;
	E2_ERR_DECLARE
#endif
#ifdef E2_VFSTMP
	E2_FSType dirtype = ?; //FIXME
#else
	E2_FSType dirtype = curr_view->dirtype; //FIXME
#endif

	//populate the local statbuf, if we can
	if ((*twdata->twstat) (local, &statbuf E2_ERR_PTR()))
	{	//failed (statbuf will be garbage for this call)

		//since cbfunc doesn't have access to the error data, as a
		//convenience we report here
		if (!twdata->exec_flags & E2TW_THREADED)
			gdk_threads_leave ();
		e2_fs_error_local (_("Cannot stat %s"), (gchar *)local E2_ERR_MSGL());
		if (!twdata->exec_flags & E2TW_THREADED)
			gdk_threads_enter ();

		result = (*twdata->cbfunc) (local, &statbuf, E2TW_NS, twdata->user_data);
#ifdef E2_VFS
		if (twdata->exec_flags & E2TW_XERR)
		{
			callerr = (GError **)twdata->user_data;
			if (callerr != NULL && *callerr == NULL)	//valid, unused error data
				*callerr = E2_ERR_NAME;
		}
		else
			E2_ERR_CLEAR
#endif
		return ((result == E2TW_CONTINUE) ? result : E2TW_STOP);	//for an un-statable item, only continue or stop
	}
	//check whether it is ok to open the dir
	newdepth = depth + 1;
	if (twdata->max_depth != -1 && newdepth > twdata->max_depth)
	{
		result = (*twdata->cbfunc) (local, &statbuf, E2TW_DL, twdata->user_data);
		return ((result == E2TW_STOP) ? result : E2TW_CONTINUE); //max depth reached, would expect normally to continue
	}

	else if ((twdata->exec_flags & E2TW_MOUNT) &&
		twdata->device != statbuf.st_dev)
	{
		result = (*twdata->cbfunc) (local, &statbuf, E2TW_DM, twdata->user_data);
		return ((result == E2TW_STOP) ? result : E2TW_CONTINUE); //unwanted filesystem, would expect normally to continue
	}
	if (twdata->exec_flags & E2TW_THREADED)
		gdk_threads_enter ();
	//get list of item-names in the dir
	entries = (GList *)e2_fs_dir_foreach ((gchar *)local, dirtype,
		 NULL, NULL, NULL E2_ERR_PTR());
	if (twdata->exec_flags & E2TW_THREADED)
		gdk_threads_leave ();
	//check whether it was possible to open the dir
	if (E2DREAD_FAILED (entries))	//or (gpointer)entries == GUINT_TO_POINTER (E2DREAD_DNR))
	{
		//since cbfunc doesn't have access to the error data, as a
		//convenience we report here
		if (!twdata->exec_flags & E2TW_THREADED)
			gdk_threads_leave ();
		e2_fs_error_local (_("Cannot open directory %s"),
				(gchar *)local E2_ERR_MSGL());
		if (!twdata->exec_flags & E2TW_THREADED)
			gdk_threads_enter ();

		result = (*twdata->cbfunc) (local, &statbuf, E2TW_DNR, twdata->user_data);
		//continue if no problem or a problem down the tree
		//any problem is handled by the relevant tw func and its continued-after-problem flag
#ifdef E2_VFS
		if (twdata->exec_flags & E2TW_XERR)
		{
			callerr = (GError **)twdata->user_data;
			if (callerr != NULL && *callerr == NULL)	//valid, unused error data
				*callerr = E2_ERR_NAME;
		}
		else
			E2_ERR_CLEAR
#endif
		return (((result == E2TW_SKIP_SUBTREE && newdepth > 1)
					|| result == E2TW_CONTINUE) ?
//		return ((result == E2TW_CONTINUE) ?
					E2TW_CONTINUE : E2TW_STOP); //continue if so ordered
	}
	else	//this dir opened successfully
	{
		//CHECKME want these flags to be variable by any callback function ?
		gboolean reportdirs = !(twdata->exec_flags & E2TW_NODIR);
		gboolean reportnondirs = !(twdata->exec_flags & E2TW_ONLYDIR);

		//if dir-start report for this dir not issued before, do so now
		if ((twdata->exec_flags & E2TW_DEPTH)	//depth-1st
			|| newdepth == 1)	//breadth-1st, but this is the 1st call
		{
			if (reportdirs)
			{
				result = (*twdata->cbfunc) (local, &statbuf, E2TW_D, twdata->user_data);

				if (result == E2TW_SKIP_SUBTREE)
				{
					e2_list_free_with_data (&entries);
					//continue if the problem is down the tree
					//any problem is handled by the relevant tw func and its continued-after-problem flag
					return ((newdepth > 1) ? E2TW_CONTINUE : E2TW_STOP);
	//				return E2TW_STOP;
				}
				else if (result == E2TW_STOP)
				{
					e2_list_free_with_data (&entries);
					return result;
				}
			}
		}

		if (twdata->exec_flags & E2TW_CHDIR)
		{
#ifdef E2_VFSTMP
		//FIXME what dir to open local may be virtual
#endif
			if (e2_fs_chdir_local (local E2_ERR_PTR()))
			{
				if (!twdata->exec_flags & E2TW_THREADED)
					gdk_threads_leave ();
				e2_fs_error_local (_("Cannot open directory %s"),
					(gchar *) local E2_ERR_MSGL());
				if (!twdata->exec_flags & E2TW_THREADED)
					gdk_threads_enter ();
#ifdef E2_VFS
				if (twdata->exec_flags & E2TW_XERR)
				{
					callerr = (GError **)twdata->user_data;
					if (callerr != NULL && *callerr == NULL)	//valid, unused error data
						*callerr = E2_ERR_NAME;
					else
						E2_ERR_CLEAR
				}
				else
					E2_ERR_CLEAR
#endif
				return E2TW_FAILED;
			}
		}

//======= DIR OPEN, CWD NOW LOCAL
		parked_dirs = NULL;
		gint len1 = strlen (local);
		gint len2, len3;
		//walk it (tree-depth now = newdepth)
		for (tmp = entries; tmp != NULL ; tmp = g_list_next (tmp))
		{
			if (strcmp ((gchar *)tmp->data, "..") == 0)
				continue;

			//construct item's full path (mimic g_strdup_printf())
			len2 = strlen ((gchar *)tmp->data);
			len3 = len1 + len2 + sizeof (gchar);
			//space for strings, with joining "/" and trailing '\0'
//			gchar *path = g_malloc ((gulong) (len1 + len2 + sizeof (gchar) + 1));
			gchar *path = g_malloc ((gulong) (len3 + 1));	//FIXME g_try_malloc
#ifdef __USE_GNU
			gchar *next = mempcpy (path, local, len1);
			if (*(next-1) != G_DIR_SEPARATOR)	//ascii check is sufficient
				next = mempcpy (next, G_DIR_SEPARATOR_S, sizeof (gchar));
			next = mempcpy (next, tmp->data, len2);
			*next = '\0';
#else
			memcpy (path, local, len1);
			if (*(path+len1-sizeof (gchar)) == G_DIR_SEPARATOR)
				len1 -= sizeof (gchar);	//fudge next copy
			else
				memcpy (path+len1, G_DIR_SEPARATOR_S, sizeof (gchar));
			memcpy (path+len1+sizeof (gchar), tmp->data, len2);
			*(path+len3) = '\0';
#endif
			if ((*twdata->twstat) (path, &statbuf E2_ERR_PTR()))
			{	//process stat failure
				//NOTE statting a hanging link to a dir does NOT fail here, glib BUG ?
				//since cbfunc doesn't have access to the error data, as a
				//convenience we report here
				if (!twdata->exec_flags & E2TW_THREADED)
					gdk_threads_leave ();
				e2_fs_error_local (_("Cannot stat %s"), path E2_ERR_MSGL());
				if (!twdata->exec_flags & E2TW_THREADED)
					gdk_threads_enter ();

				result = (*twdata->cbfunc) (path, &statbuf, E2TW_NS, twdata->user_data);
#ifdef E2_VFS
				if (twdata->exec_flags & E2TW_XERR)
				{
					callerr = (GError **)twdata->user_data;
					if (callerr != NULL && *callerr == NULL)	//valid, unused error data
						*callerr = E2_ERR_NAME;
				}
				else
					E2_ERR_CLEAR
#endif
				if (result == E2TW_STOP)
				{
					if (! (twdata->exec_flags & E2TW_DEPTH)
						&& parked_dirs != NULL)
							e2_list_free_with_data (&parked_dirs);
					e2_list_free_with_data (&entries);
					g_free (path);
					return result;
				}
			}
			else if (S_ISDIR (statbuf.st_mode))
			{	//stattable dir or (present) link-target that's a dir
				if (!(twdata->exec_flags & E2TW_DEPTH))
				{	//breadth-first walk
					if (reportdirs)
					{
						//issue the dir-start report here
						result = (*twdata->cbfunc) (path, &statbuf, E2TW_D, twdata->user_data);

						if (result == E2TW_SKIP_SUBTREE)
							result = E2TW_STOP;
						if (result == E2TW_STOP)
						{
							e2_list_free_with_data (&entries);
							g_free (path);
							return result;
						}
					}
					//log the dir, for later opening
					parked_dirs = g_list_append (parked_dirs, strdup (path));
				}
				else
				{	//depth-first walk, open the dir now
					//(start- and end-reports issued in the recursed function)
					if (_e2_fs_tw (path, newdepth, twdata) != E2TW_CONTINUE)
					{
						//no parks to cleanup if depth-1st walk
						e2_list_free_with_data (&entries);
						g_free (path);
						return E2TW_STOP;
					}
					//restore system dir to where we are
					if (twdata->exec_flags & E2TW_CHDIR)
					{
#ifdef E2_VFSTMP
		//FIXME what dir to open local may be virtual
#endif
						if (e2_fs_chdir_local (local E2_ERR_PTR()))
						{
							if (!twdata->exec_flags & E2TW_THREADED)
								gdk_threads_leave ();
							e2_fs_error_local (_("Cannot open directory %s"),
								(gchar *)local E2_ERR_MSGL());
							if (!twdata->exec_flags & E2TW_THREADED)
								gdk_threads_enter ();
#ifdef E2_VFS
							if (twdata->exec_flags & E2TW_XERR)
							{
								callerr = (GError **)twdata->user_data;
								if (callerr != NULL && *callerr == NULL)	//valid, unused error data
									*callerr = E2_ERR_NAME;
								else
									E2_ERR_CLEAR
							}
							else
								E2_ERR_CLEAR
#endif
							return E2TW_FAILED;
						}
					}
				}
			}
			else	//not a dir (it could be a hanging link to a dir)
			{
				if (reportnondirs &&
					_e2_fs_tw_nondir (path, &statbuf, twdata) != E2TW_CONTINUE)
				{
					if (! (twdata->exec_flags & E2TW_DEPTH)
						&& parked_dirs != NULL)
							e2_list_free_with_data (&parked_dirs);
					e2_list_free_with_data (&entries);
					g_free (path);
					return E2TW_STOP;
				}
			}
			g_free (path);
		}
		e2_list_free_with_data (&entries);
//========= DIR CLOSED, CWD STILL LOCAL
	}

	if (!(twdata->exec_flags & E2TW_DEPTH) && parked_dirs != NULL)
	{	//breadth-first walk
		//process the dirs that haven't yet been opened
		for (tmp = parked_dirs ; tmp != NULL ; tmp = g_list_next (tmp))
		{
			gchar *path = (gchar *) tmp->data;
			//remember ! start-report already issued, do not repeat inside the recurse
			if (_e2_fs_tw (path, newdepth, twdata) != E2TW_CONTINUE)
			{
				e2_list_free_with_data (&parked_dirs);
				//CHECKME dir end-report here, if aborted ?
				return E2TW_STOP;
			}
		}
		e2_list_free_with_data (&parked_dirs);
	}

	if (!(*twdata->twstat) (local, &statbuf E2_ERR_PTR()))	//freshen the statbuf FIXME another copy instead of refresh ??
	{
		//issue dir-end report
		result = (*twdata->cbfunc) (local, &statbuf, E2TW_DP, twdata->user_data);
		return (result == E2TW_STOP) ? result : E2TW_CONTINUE;
	}
	else
	{
		//can't issue valid dir-end report
#ifdef E2_VFS
//FIXME warn user about error if not already done downstream
		if (twdata->exec_flags & E2TW_XERR)
		{
			callerr = (GError **)twdata->user_data;
			if (callerr != NULL && *callerr == NULL)	//valid, unused error data
				*callerr = E2_ERR_NAME;
		}
		else
			E2_ERR_CLEAR
#endif
		return E2TW_STOP;
	}
}
/**
 * @brief locate and report all items in directory tree under @a start_dir
 *
 * This is a replacement for nftw(), with user data (to avoid global variables
 * in callback functions, so threads can be used) and extra types of status-
 * report provided to callback functions, and to enable vfs support
 * Potential error message expects gtk's BGL to be closed, unless flag
 * E2TW_THREADED is set. That flag is not passed to downstream callbacks, which
 * must therefore manage mutexes themselves.
 *
 * @param start_item localised string, absolute path of item (dir or non-dir) in which to start the walk
 * @param callback pointer to the function to be called for each located item
 * @param user_data pointer to user-specified data to be supplied to each call to @a callback
 * @param max_depth the largest number of tree levels to descend, or -1 for no limit
 * @param exec_flags flags indicating the way the walk is to be performed
 *
 * @return TRUE if all of the tree was processed in way(s) that @a callback considered ok
 **/
gboolean e2_fs_tw (gchar *start_item, E2_TwResult (*callback) (),
	void *user_data, gint max_depth, E2_TwFlags exec_flags E2_ERR_ARG())
{
	E2_TwResult result;
	struct stat statbuf;
#ifdef E2_VFS
	GError **callerr;
#endif

	//NOTE: not lstat(), because downstream _e2_fs_tw_nondir() needs to follow a stat()
	if (e2_fs_stat (start_item, &statbuf E2_ERR_SAMEARG()))
	{
		if (!exec_flags & E2TW_THREADED)
			gdk_threads_leave ();
		e2_fs_error_local (_("Cannot stat %s"), start_item E2_ERR_MSGC());
		if (!exec_flags & E2TW_THREADED)
			gdk_threads_enter ();
#ifdef E2_VFS
		if (exec_flags & E2TW_XERR)
		{
			callerr = (GError **)user_data;
			if (callerr != NULL && *callerr == NULL)	//valid, unused error data
				*callerr = *E2_ERR_NAME;
		}
#endif
		return FALSE;
	}

	//park the walk-data where downstream functions can get it
	E2_TwData data;
	data.twstat = (exec_flags & E2TW_PHYS) ? e2_fs_lstat : e2_fs_stat;
	data.cbfunc = callback;
	data.user_data = user_data;
	data.max_depth = max_depth;
	data.exec_flags = exec_flags;
//	if (exec_flags & E2TW_MOUNT)
//	{
		data.device = statbuf.st_dev;
//	}

	if (!S_ISDIR (statbuf.st_mode))
	{	//selected item not a dir (or a link to one)
		result = _e2_fs_tw_nondir (start_item, &statbuf, &data);
	}
	else //it's a dir ...
	{
		gchar *original_path;

		if (exec_flags & E2TW_CHDIR)
		{
			gchar *first_path;
//			gchar *s = NULL;
//			if (S_ISDIR (statbuf.st_mode))
#ifdef E2_VFSTMP
	//FIXME which dir
#else
				first_path = start_item;
#endif
/*			else
			{
				s = strrchr (start_item, G_DIR_SEPARATOR);
				if (s != start_item)
					*s = '\0';
				else
				{
					first_path = G_DIR_SEPARATOR_S;
					s = NULL;
				}
			}
*/
#ifdef E2_VFSTMP
	//FIXME which CWD to save
	//FIXME what dir to open local may be virtual
#else
			original_path = getcwd (NULL, 0);	//CHECKME error test ?

			if (e2_fs_chdir_local (first_path E2_ERR_SAMEARG()))
#endif
			{
				if (!exec_flags & E2TW_THREADED)
					gdk_threads_leave ();
				e2_fs_error_local (_("Cannot open directory %s"),
					first_path E2_ERR_MSGC());
				if (!exec_flags & E2TW_THREADED)
					gdk_threads_enter ();
#ifdef E2_VFS
				if (exec_flags & E2TW_XERR)
				{
					callerr = (GError **)user_data;
					if (callerr != NULL && *callerr == NULL)	//valid, unused error data
						*callerr = *E2_ERR_NAME;
				}
#endif
				g_free (original_path);
				return FALSE;
			}
//			if (s != NULL)
//				*s = G_DIR_SEPARATOR;
		}

			result = _e2_fs_tw (start_item, 0, &data);	//DO IT
#ifdef E2_VFS
			if (exec_flags & E2TW_XERR)
			{
				callerr = (GError **)user_data;
				if (callerr != NULL && *callerr != NULL)	//error data recorded
					*E2_ERR_NAME = *callerr;
			}
#endif

		if (exec_flags & E2TW_CHDIR)
		{
#ifdef E2_VFSTMP
		//FIXME what dir to open local may be virtual
#endif
			if (e2_fs_chdir_local (original_path E2_ERR_SAMEARG()))
			{
				if (!exec_flags & E2TW_THREADED)
					gdk_threads_leave ();
				e2_fs_error_local (_("Cannot open directory %s"),
					original_path E2_ERR_MSGC());
				if (!exec_flags & E2TW_THREADED)
					gdk_threads_enter ();
#ifdef E2_VFS
				if (exec_flags & E2TW_XERR)
				{
					callerr = (GError **)user_data;
					if (callerr != NULL && *callerr == NULL)	//valid, unused error data
						*callerr = *E2_ERR_NAME;
				}
#endif
/* no need for a compelled failure at end of walk
				g_free (original_path);
				return FALSE;
*/
			}
			g_free (original_path);
		}
	}

	return (result == E2TW_CONTINUE);
}
